@@ -3,7 +3,7 @@
  * Copyright (c) 2004-2005 The Trustees of Indiana University and Indiana
  *                         University Research and Technology
  *                         Corporation.  All rights reserved.
- * Copyright (c) 2004-2005 The University of Tennessee and The University
+ * Copyright (c) 2004-2020 The University of Tennessee and The University
  *                         of Tennessee Research Foundation.  All rights
  *                         reserved.
  * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart,
@@ -13,12 +13,15 @@
  * Copyright (c) 2007      Lawrence Livermore National Security, LLC.  All
  *                         rights reserved.
  * Copyright (c) 2008      Sun Microsystems, Inc.  All rights reserved.
+ * Copyright (c) 2010-2012 Oak Ridge National Laboratory.  All rights reserved.
  * Copyright (c) 2008-2014 Cisco Systems, Inc.  All rights reserved.
- * Copyright (c) 2012      Oak Ridge National Laboratory.  All rights reserved.
  * Copyright (c) 2013      Los Alamos National Security, LLC. All rights
  *                         reserved.
  * Copyright (c) 2014      Research Organization for Information Science
  *                         and Technology (RIST). All rights reserved.
+ * Copyright (c) 2016-2017 IBM Corporation.  All rights reserved.
+ * Copyright (c) 2017      FUJITSU LIMITED.  All rights reserved.
+ * Copyright (c) 2020      BULL S.A.S. All rights reserved.
  * $COPYRIGHT$
  *
  * Additional copyrights may follow
@@ -35,6 +38,7 @@
 #include "mpi.h"
 #include "ompi/communicator/communicator.h"
 #include "opal/util/output.h"
+#include "opal/util/argv.h"
 #include "opal/util/show_help.h"
 #include "opal/class/opal_list.h"
 #include "opal/class/opal_object.h"
@@ -42,19 +46,13 @@
 #include "opal/mca/base/base.h"
 #include "ompi/mca/coll/coll.h"
 #include "ompi/mca/coll/base/base.h"
-
+#include "ompi/mca/coll/base/coll_base_util.h"
+#include "ompi/mca/coll/base/coll_base_functions.h"
 
 /*
- * Local types
+ * Stuff for the OBJ interface
  */
-struct avail_coll_t {
-    opal_list_item_t super;
-
-    int ac_priority;
-    mca_coll_base_module_2_1_0_t *ac_module;
-};
-typedef struct avail_coll_t avail_coll_t;
-
+OBJ_CLASS_INSTANCE(mca_coll_base_avail_coll_t, opal_list_item_t, NULL, NULL);
 
 /*
  * Local functions
@@ -63,37 +61,31 @@ static opal_list_t *check_components(opal_list_t * components,
                                      ompi_communicator_t * comm);
 static int check_one_component(ompi_communicator_t * comm,
                                const mca_base_component_t * component,
-                               mca_coll_base_module_2_1_0_t ** module);
+                               mca_coll_base_module_t ** module);
 
 static int query(const mca_base_component_t * component,
                  ompi_communicator_t * comm, int *priority,
-                 mca_coll_base_module_2_1_0_t ** module);
+                 mca_coll_base_module_t ** module);
 
-static int query_2_0_0(const mca_coll_base_component_2_0_0_t *
+static int query_2_4_0(const mca_coll_base_component_2_4_0_t *
                        coll_component, ompi_communicator_t * comm,
                        int *priority,
-                       mca_coll_base_module_2_1_0_t ** module);
-
-/*
- * Stuff for the OBJ interface
- */
-static OBJ_CLASS_INSTANCE(avail_coll_t, opal_list_item_t, NULL, NULL);
-
+                       mca_coll_base_module_t ** module);
 
 #define COPY(module, comm, func)                                        \
     do {                                                                \
         if (NULL != module->coll_ ## func) {                            \
-            if (NULL != comm->c_coll.coll_ ## func ## _module) {        \
-                OBJ_RELEASE(comm->c_coll.coll_ ## func ## _module);     \
+            if (NULL != comm->c_coll->coll_ ## func ## _module) {       \
+                OBJ_RELEASE(comm->c_coll->coll_ ## func ## _module);    \
             }                                                           \
-            comm->c_coll.coll_ ## func = module->coll_ ## func;         \
-            comm->c_coll.coll_ ## func ## _module = module;             \
+            comm->c_coll->coll_ ## func = module->coll_ ## func;        \
+            comm->c_coll->coll_ ## func ## _module = module;            \
             OBJ_RETAIN(module);                                         \
         }                                                               \
     } while (0)
 
 #define CHECK_NULL(what, comm, func)                                    \
-  ( (what) = # func , NULL == (comm)->c_coll.coll_ ## func)
+  ( (what) = # func , NULL == (comm)->c_coll->coll_ ## func)
 
 /*
  * This function is called at the initialization time of every
@@ -110,13 +102,14 @@ int mca_coll_base_comm_select(ompi_communicator_t * comm)
     int ret;
 
     /* Announce */
-    opal_output_verbose(10, ompi_coll_base_framework.framework_output,
-                        "coll:base:comm_select: new communicator: %s (cid %d)",
-                        comm->c_name, comm->c_contextid);
+    opal_output_verbose(9, ompi_coll_base_framework.framework_output,
+                        "coll:base:comm_select: new communicator: %s (cid %s)",
+                        comm->c_name, ompi_comm_print_cid (comm));
 
     /* Initialize all the relevant pointers, since they're used as
      * sentinel values */
-    memset(&comm->c_coll, 0, sizeof(mca_coll_base_comm_coll_t));
+    comm->c_coll = (mca_coll_base_comm_coll_t*)calloc(1, sizeof(mca_coll_base_comm_coll_t));
+    comm->c_coll->coll_revoke_local = mca_coll_base_revoke_local;
 
     opal_output_verbose(10, ompi_coll_base_framework.framework_output,
                         "coll:base:comm_select: Checking all available modules");
@@ -135,15 +128,27 @@ int mca_coll_base_comm_select(ompi_communicator_t * comm)
     /* FIX ME - Do some kind of collective operation to find a module
        that everyone has available */
 
+    /* List to store every valid module */
+    comm->c_coll->module_list =  OBJ_NEW(opal_list_t);
+
     /* do the selection loop */
     for (item = opal_list_remove_first(selectable);
          NULL != item; item = opal_list_remove_first(selectable)) {
 
-        avail_coll_t *avail = (avail_coll_t *) item;
+        mca_coll_base_avail_coll_t *avail = (mca_coll_base_avail_coll_t *) item;
 
         /* initialize the module */
         ret = avail->ac_module->coll_module_enable(avail->ac_module, comm);
+
+        opal_output_verbose(9, ompi_coll_base_framework.framework_output,
+                            "coll:base:comm_select: selecting  %10s, priority %3d, %s",
+                            avail->ac_component_name, avail->ac_priority,
+                            (OMPI_SUCCESS == ret ? "Enabled": "Disabled") );
+
         if (OMPI_SUCCESS == ret) {
+            /* Save every component that is initialized,
+             * queried and enabled successfully */
+            opal_list_append(comm->c_coll->module_list, &avail->super);
 
             /* copy over any of the pointers */
             COPY(avail->ac_module, comm, allgather);
@@ -182,6 +187,24 @@ int mca_coll_base_comm_select(ompi_communicator_t * comm)
             COPY(avail->ac_module, comm, iscatter);
             COPY(avail->ac_module, comm, iscatterv);
 
+            COPY(avail->ac_module, comm, allgather_init);
+            COPY(avail->ac_module, comm, allgatherv_init);
+            COPY(avail->ac_module, comm, allreduce_init);
+            COPY(avail->ac_module, comm, alltoall_init);
+            COPY(avail->ac_module, comm, alltoallv_init);
+            COPY(avail->ac_module, comm, alltoallw_init);
+            COPY(avail->ac_module, comm, barrier_init);
+            COPY(avail->ac_module, comm, bcast_init);
+            COPY(avail->ac_module, comm, exscan_init);
+            COPY(avail->ac_module, comm, gather_init);
+            COPY(avail->ac_module, comm, gatherv_init);
+            COPY(avail->ac_module, comm, reduce_init);
+            COPY(avail->ac_module, comm, reduce_scatter_block_init);
+            COPY(avail->ac_module, comm, reduce_scatter_init);
+            COPY(avail->ac_module, comm, scan_init);
+            COPY(avail->ac_module, comm, scatter_init);
+            COPY(avail->ac_module, comm, scatterv_init);
+
             /* We can not reliably check if this comm has a topology
              * at this time. The flags are set *after* coll_select */
             COPY(avail->ac_module, comm, neighbor_allgather);
@@ -195,10 +218,24 @@ int mca_coll_base_comm_select(ompi_communicator_t * comm)
             COPY(avail->ac_module, comm, ineighbor_alltoall);
             COPY(avail->ac_module, comm, ineighbor_alltoallv);
             COPY(avail->ac_module, comm, ineighbor_alltoallw);
+
+            COPY(avail->ac_module, comm, neighbor_allgather_init);
+            COPY(avail->ac_module, comm, neighbor_allgatherv_init);
+            COPY(avail->ac_module, comm, neighbor_alltoall_init);
+            COPY(avail->ac_module, comm, neighbor_alltoallv_init);
+            COPY(avail->ac_module, comm, neighbor_alltoallw_init);
+
+            COPY(avail->ac_module, comm, reduce_local);
+
+#if OPAL_ENABLE_FT_MPI
+            COPY(avail->ac_module, comm, agree);
+            COPY(avail->ac_module, comm, iagree);
+#endif
+        } else {
+            /* release the original module reference and the list item */
+            OBJ_RELEASE(avail->ac_module);
+            OBJ_RELEASE(avail);
         }
-        /* release the original module reference and the list item */
-        OBJ_RELEASE(avail->ac_module);
-        OBJ_RELEASE(avail);
     }
 
     /* Done with the list from the check_components() call so release it. */
@@ -238,7 +275,25 @@ int mca_coll_base_comm_select(ompi_communicator_t * comm)
         CHECK_NULL(which_func, comm, ireduce_scatter) ||
         ((OMPI_COMM_IS_INTRA(comm)) && CHECK_NULL(which_func, comm, iscan)) ||
         CHECK_NULL(which_func, comm, iscatter) ||
-        CHECK_NULL(which_func, comm, iscatterv)) {
+        CHECK_NULL(which_func, comm, iscatterv) ||
+        CHECK_NULL(which_func, comm, allgather_init) ||
+        CHECK_NULL(which_func, comm, allgatherv_init) ||
+        CHECK_NULL(which_func, comm, allreduce_init) ||
+        CHECK_NULL(which_func, comm, alltoall_init) ||
+        CHECK_NULL(which_func, comm, alltoallv_init) ||
+        CHECK_NULL(which_func, comm, alltoallw_init) ||
+        CHECK_NULL(which_func, comm, barrier_init) ||
+        CHECK_NULL(which_func, comm, bcast_init) ||
+        ((OMPI_COMM_IS_INTRA(comm)) && CHECK_NULL(which_func, comm, exscan_init)) ||
+        CHECK_NULL(which_func, comm, gather_init) ||
+        CHECK_NULL(which_func, comm, gatherv_init) ||
+        CHECK_NULL(which_func, comm, reduce_init) ||
+        CHECK_NULL(which_func, comm, reduce_scatter_block_init) ||
+        CHECK_NULL(which_func, comm, reduce_scatter_init) ||
+        ((OMPI_COMM_IS_INTRA(comm)) && CHECK_NULL(which_func, comm, scan_init)) ||
+        CHECK_NULL(which_func, comm, scatter_init) ||
+        CHECK_NULL(which_func, comm, scatterv_init) ||
+        CHECK_NULL(which_func, comm, reduce_local) ) {
         /* TODO -- Once the topology flags are set before coll_select then
          * check if neighborhood collectives have been set. */
 
@@ -253,8 +308,8 @@ int mca_coll_base_comm_select(ompi_communicator_t * comm)
 
 static int avail_coll_compare (opal_list_item_t **a,
                                opal_list_item_t **b) {
-    avail_coll_t *acoll = (avail_coll_t *) *a;
-    avail_coll_t *bcoll = (avail_coll_t *) *b;
+    mca_coll_base_avail_coll_t *acoll = (mca_coll_base_avail_coll_t *) *a;
+    mca_coll_base_avail_coll_t *bcoll = (mca_coll_base_avail_coll_t *) *b;
 
     if (acoll->ac_priority > bcoll->ac_priority) {
         return 1;
@@ -265,6 +320,20 @@ static int avail_coll_compare (opal_list_item_t **a,
     return 0;
 }
 
+static inline int
+component_in_argv(char **argv, const char* component_name)
+{
+    if( NULL != argv ) {
+        while( NULL != *argv ) {
+            if( 0 == strcmp(component_name, *argv) ) {
+                return 1;
+            }
+            argv++;  /* move to the next argument */
+        }
+    }
+    return 0;
+}
+
 /*
  * For each module in the list, check and see if it wants to run, and
  * do the resulting priority comparison.  Make a list of modules to be
@@ -274,13 +343,60 @@ static int avail_coll_compare (opal_list_item_t **a,
 static opal_list_t *check_components(opal_list_t * components,
                                      ompi_communicator_t * comm)
 {
-    int priority;
+    int priority, flag;
+    int count_include = 0;
     const mca_base_component_t *component;
     mca_base_component_list_item_t *cli;
-    mca_coll_base_module_2_1_0_t *module;
+    mca_coll_base_module_t *module;
     opal_list_t *selectable;
-    avail_coll_t *avail;
+    mca_coll_base_avail_coll_t *avail;
+    char **coll_argv = NULL, **coll_exclude = NULL, **coll_include = NULL;
 
+    /* Check if this communicator comes with restrictions on the collective modules
+     * it wants to use. The restrictions are consistent with the MCA parameter
+     * to limit the collective components loaded, but it applies for each
+     * communicator and is provided as an info key during the communicator
+     * creation. Unlike the MCA param, this info key is used not to select
+     * components but either to prevent components from being used or to
+     * force a change in the component priority.
+     */
+    if( NULL != comm->super.s_info) {
+        opal_cstring_t *info_str;
+        opal_info_get(comm->super.s_info, "ompi_comm_coll_preference",
+                      &info_str, &flag);
+        if( !flag ) {
+            goto proceed_to_select;
+        }
+        coll_argv = opal_argv_split(info_str->string, ',');
+        OBJ_RELEASE(info_str);
+        if(NULL == coll_argv) {
+            goto proceed_to_select;
+        }
+        int idx2;
+        count_include = opal_argv_count(coll_argv);
+        /* Allocate the coll_include argv */
+        coll_include = (char**)malloc((count_include + 1) * sizeof(char*));
+        coll_include[count_include] = NULL; /* NULL terminated array */
+        /* Dispatch the include/exclude in the corresponding arrays */
+        for( int idx = 0; NULL != coll_argv[idx]; idx++ ) {
+            if( '^' == coll_argv[idx][0] ) {
+                coll_include[idx] = NULL;  /* NULL terminated array */
+
+                /* Allocate the coll_exclude argv */
+                coll_exclude = (char**)malloc((count_include - idx + 1) * sizeof(char*));
+                /* save the exclude components */
+                for( idx2 = idx; NULL != coll_argv[idx2]; idx2++ ) {
+                    coll_exclude[idx2 - idx] = coll_argv[idx2];
+                }
+                coll_exclude[idx2 - idx] = NULL;  /* NULL-terminated array */
+                coll_exclude[0] = coll_exclude[0] + 1;  /* get rid of the ^ */
+                count_include = idx;
+                break;
+            }
+            coll_include[idx] = coll_argv[idx];
+        }
+    }
+ proceed_to_select:
     /* Make a list of the components that query successfully */
     selectable = OBJ_NEW(opal_list_t);
 
@@ -288,27 +404,78 @@ static opal_list_t *check_components(opal_list_t * components,
     OPAL_LIST_FOREACH(cli, &ompi_coll_base_framework.framework_components, mca_base_component_list_item_t) {
         component = cli->cli_component;
 
+        /* dont bother is we have this component in the exclusion list */
+        if( component_in_argv(coll_exclude, component->mca_component_name) ) {
+            opal_output_verbose(10, ompi_coll_base_framework.framework_output,
+                                "coll:base:comm_select: component disqualified: %s (due to communicator info key)",
+                                component->mca_component_name );
+            continue;
+        }
         priority = check_one_component(comm, component, &module);
         if (priority >= 0) {
             /* We have a component that indicated that it wants to run
                by giving us a module */
-            avail = OBJ_NEW(avail_coll_t);
+            avail = OBJ_NEW(mca_coll_base_avail_coll_t);
             avail->ac_priority = priority;
             avail->ac_module = module;
+            // Point to the string so we don't have to free later
+            avail->ac_component_name = component->mca_component_name;
 
             opal_list_append(selectable, &avail->super);
         }
+        else {
+            opal_output_verbose(10, ompi_coll_base_framework.framework_output,
+                                "coll:base:comm_select: component disqualified: %s (priority %d < 0)",
+                                component->mca_component_name, priority );
+
+            // If the disqualified collective returned a module make sure we
+            // release it here, since it will become a leak otherwise.
+            if( NULL != module ) {
+                OBJ_RELEASE(module);
+                module = NULL;
+            }
+        }
     }
 
     /* If we didn't find any available components, return an error */
     if (0 == opal_list_get_size(selectable)) {
         OBJ_RELEASE(selectable);
+        if( NULL != coll_exclude ) {
+            free(coll_exclude);
+        }
+        if( NULL != coll_include ) {
+            free(coll_include);
+        }
         return NULL;
     }
 
     /* Put this list in priority order */
     opal_list_sort(selectable, avail_coll_compare);
 
+    /* For all valid component reorder them not on their provided priorities but on
+     * the order requested in the info key. As at this point the coll_include is
+     * already ordered backward we can simply append the components.
+     * Note that the last element in selectable will have the highest priority.
+     */
+    for (int idx = count_include-1; idx >= 0; --idx) {
+        mca_coll_base_avail_coll_t *item;
+        OPAL_LIST_FOREACH(item, selectable, mca_coll_base_avail_coll_t) {
+            if (0 == strcmp(item->ac_component_name, coll_include[idx])) {
+                opal_list_remove_item(selectable, &item->super);
+                opal_list_append(selectable, &item->super);
+                break;
+            }
+        }
+    }
+
+    opal_argv_free(coll_argv);
+    if( NULL != coll_exclude ) {
+        free(coll_exclude);
+    }
+    if( NULL != coll_include ) {
+        free(coll_include);
+    }
+
     /* All done */
     return selectable;
 }
@@ -319,7 +486,7 @@ static opal_list_t *check_components(opal_list_t * components,
  */
 static int check_one_component(ompi_communicator_t * comm,
                                const mca_base_component_t * component,
-                               mca_coll_base_module_2_1_0_t ** module)
+                               mca_coll_base_module_t ** module)
 {
     int err;
     int priority = -1;
@@ -342,7 +509,6 @@ static int check_one_component(ompi_communicator_t * comm,
     return priority;
 }
 
-
 /**************************************************************************
  * Query functions
  **************************************************************************/
@@ -353,16 +519,14 @@ static int check_one_component(ompi_communicator_t * comm,
  */
 static int query(const mca_base_component_t * component,
                  ompi_communicator_t * comm,
-                 int *priority, mca_coll_base_module_2_1_0_t ** module)
+                 int *priority, mca_coll_base_module_t ** module)
 {
     *module = NULL;
     if (2 == component->mca_type_major_version &&
-        0 == component->mca_type_minor_version &&
+        4 == component->mca_type_minor_version &&
         0 == component->mca_type_release_version) {
-        const mca_coll_base_component_2_0_0_t *coll100 =
-            (mca_coll_base_component_2_0_0_t *) component;
 
-        return query_2_0_0(coll100, comm, priority, module);
+        return query_2_4_0((const mca_coll_base_component_2_4_0_t *)component, comm, priority, module);
     }
 
     /* Unknown coll API version -- return error */
@@ -371,11 +535,11 @@ static int query(const mca_base_component_t * component,
 }
 
 
-static int query_2_0_0(const mca_coll_base_component_2_0_0_t * component,
+static int query_2_4_0(const mca_coll_base_component_2_4_0_t * component,
                        ompi_communicator_t * comm, int *priority,
-                       mca_coll_base_module_2_1_0_t ** module)
+                       mca_coll_base_module_t ** module)
 {
-    mca_coll_base_module_2_1_0_t *ret;
+    mca_coll_base_module_t *ret;
 
     /* There's currently no need for conversion */
 
