@@ -1,8 +1,9 @@
 /*
- * Copyright (c) 2014      The University of Tennessee and The University
+ * Copyright (c) 2014-2017 The University of Tennessee and The University
  *                         of Tennessee Research Foundation.  All rights
  *                         reserved.
  * Copyright (c) 2014-2015 NVIDIA Corporation.  All rights reserved.
+ * Copyright (c) 2022      Amazon.com, Inc. or its affiliates.  All Rights reserved.
  * $COPYRIGHT$
  *
  * Additional copyrights may follow
@@ -17,7 +18,6 @@
 
 #include "ompi/op/op.h"
 #include "opal/datatype/opal_convertor.h"
-#include "opal/datatype/opal_datatype_cuda.h"
 
 /*
  *	allreduce_intra
@@ -34,34 +34,37 @@ mca_coll_cuda_allreduce(const void *sbuf, void *rbuf, int count,
                         mca_coll_base_module_t *module)
 {
     mca_coll_cuda_module_t *s = (mca_coll_cuda_module_t*) module;
-    ptrdiff_t true_lb, true_extent, lb, extent;
+    ptrdiff_t gap;
     char *rbuf1 = NULL, *sbuf1 = NULL, *rbuf2 = NULL;
-    const char *sbuf2;
     size_t bufsize;
     int rc;
 
-    ompi_datatype_get_extent(dtype, &lb, &extent);
-    ompi_datatype_get_true_extent(dtype, &true_lb, &true_extent);
-    bufsize = true_extent + (ptrdiff_t)(count - 1) * extent;
-    if ((MPI_IN_PLACE != sbuf) && (opal_cuda_check_bufs((char *)sbuf, NULL))) {
+    bufsize = opal_datatype_span(&dtype->super, count, &gap);
+    rc = mca_coll_cuda_check_buf((void *)sbuf);
+    if (rc < 0) {
+        return rc;
+    }
+    if ((MPI_IN_PLACE != sbuf) && (rc > 0)) {
         sbuf1 = (char*)malloc(bufsize);
         if (NULL == sbuf1) {
             return OMPI_ERR_OUT_OF_RESOURCE;
         }
-        opal_cuda_memcpy_sync(sbuf1, sbuf, bufsize);
-        sbuf2 = sbuf; /* save away original buffer */
-        sbuf = sbuf1 - true_lb;
+        mca_coll_cuda_memcpy(sbuf1, sbuf, bufsize);
+        sbuf = sbuf1 - gap;
     }
-
-    if (opal_cuda_check_bufs(rbuf, NULL)) {
+    rc = mca_coll_cuda_check_buf(rbuf);
+    if (rc < 0) {
+        return rc;
+    }
+    if (rc > 0) {
         rbuf1 = (char*)malloc(bufsize);
         if (NULL == rbuf1) {
             if (NULL != sbuf1) free(sbuf1);
             return OMPI_ERR_OUT_OF_RESOURCE;
         }
-        opal_cuda_memcpy_sync(rbuf1, rbuf, bufsize);
+        mca_coll_cuda_memcpy(rbuf1, rbuf, bufsize);
         rbuf2 = rbuf; /* save away original buffer */
-        rbuf = rbuf1 - true_lb;
+        rbuf = rbuf1 - gap;
     }
     rc = s->c_coll.coll_allreduce(sbuf, rbuf, count, dtype, op, comm, s->c_coll.coll_allreduce_module);
     if (NULL != sbuf1) {
@@ -69,7 +72,7 @@ mca_coll_cuda_allreduce(const void *sbuf, void *rbuf, int count,
     }
     if (NULL != rbuf1) {
         rbuf = rbuf2;
-        opal_cuda_memcpy_sync(rbuf, rbuf1, bufsize);
+        mca_coll_cuda_memcpy(rbuf, rbuf1, bufsize);
         free(rbuf1);
     }
     return rc;
