@@ -1,10 +1,15 @@
 /*
  * Copyright (C) 2001-2011 Mellanox Technologies Ltd. 2001-2011.  ALL RIGHTS RESERVED.
- * Copyright (c) 2016      The University of Tennessee and The University
+ * Copyright (c) 2016-2021 The University of Tennessee and The University
  *                         of Tennessee Research Foundation.  All rights
  *                         reserved.
- * Copyright (c) 2018      Research Organization for Information Science
+ * Copyright (c) 2018-2019 Research Organization for Information Science
  *                         and Technology (RIST).  All rights reserved.
+ * Copyright (c) 2018-2022 IBM Corporation.  All rights reserved.
+ * Copyright (c) 2019      Intel, Inc.  All rights reserved.
+ * Copyright (c) 2022      Amazon.com, Inc. or its affiliates.
+ *                         All Rights reserved.
+ * Copyright (c) 2026      Triad National Security, LLC. All rights reserved.
  * $COPYRIGHT$
  *
  * Additional copyrights may follow
@@ -15,65 +20,74 @@
 #include "pml_ucx.h"
 
 #include "opal/runtime/opal.h"
-#include "opal/mca/pmix/pmix.h"
+#include "opal/mca/pmix/pmix-internal.h"
 #include "ompi/attribute/attribute.h"
 #include "ompi/message/message.h"
+#include "ompi/runtime/ompi_spc.h"
 #include "ompi/mca/pml/base/pml_base_bsend.h"
 #include "opal/mca/common/ucx/common_ucx.h"
 #include "pml_ucx_request.h"
 
 #include <inttypes.h>
 
+/*
+ * Maximum number of transport entries that we can expect from UCX
+ */
+#define PML_UCX_MAX_TRANSPORT_ENTRIES 100
 
 #define PML_UCX_TRACE_SEND(_msg, _buf, _count, _datatype, _dst, _tag, _mode, _comm, ...) \
     PML_UCX_VERBOSE(8, _msg " buf %p count %zu type '%s' dst %d tag %d mode %s comm %d '%s'", \
                     __VA_ARGS__, \
                     (_buf), (_count), (_datatype)->name, (_dst), (_tag), \
-                    mca_pml_ucx_send_mode_name(_mode), (_comm)->c_contextid, \
+                    mca_pml_ucx_send_mode_name(_mode), (_comm)->c_index, \
                     (_comm)->c_name);
 
 #define PML_UCX_TRACE_RECV(_msg, _buf, _count, _datatype, _src, _tag, _comm, ...) \
     PML_UCX_VERBOSE(8, _msg " buf %p count %zu type '%s' src %d tag %d comm %d '%s'", \
                     __VA_ARGS__, \
                     (_buf), (_count), (_datatype)->name, (_src), (_tag), \
-                    (_comm)->c_contextid, (_comm)->c_name);
+                    (_comm)->c_index, (_comm)->c_name);
 
 #define PML_UCX_TRACE_PROBE(_msg, _src, _tag, _comm) \
     PML_UCX_VERBOSE(8, _msg " src %d tag %d comm %d '%s'", \
-                    _src, (_tag), (_comm)->c_contextid, (_comm)->c_name);
+                    _src, (_tag), (_comm)->c_index, (_comm)->c_name);
 
 #define PML_UCX_TRACE_MRECV(_msg, _buf, _count, _datatype, _message) \
     PML_UCX_VERBOSE(8, _msg " buf %p count %zu type '%s' msg *%p=%p (%p)", \
                     (_buf), (_count), (_datatype)->name, (void*)(_message), \
                     (void*)*(_message), (*(_message))->req_ptr);
 
+mca_pml_transports_t *mca_pml_ucx_get_transports(ompi_communicator_t *comm,
+                                                 int rank);
+
 #define MODEX_KEY "pml-ucx"
 
 mca_pml_ucx_module_t ompi_pml_ucx = {
     .super = {
-        .pml_add_procs     = mca_pml_ucx_add_procs,
-        .pml_del_procs     = mca_pml_ucx_del_procs,
-        .pml_enable        = mca_pml_ucx_enable,
-        .pml_progress      = NULL,
-        .pml_add_comm      = mca_pml_ucx_add_comm,
-        .pml_del_comm      = mca_pml_ucx_del_comm,
-        .pml_irecv_init    = mca_pml_ucx_irecv_init,
-        .pml_irecv         = mca_pml_ucx_irecv,
-        .pml_recv          = mca_pml_ucx_recv,
-        .pml_isend_init    = mca_pml_ucx_isend_init,
-        .pml_isend         = mca_pml_ucx_isend,
-        .pml_send          = mca_pml_ucx_send,
-        .pml_iprobe        = mca_pml_ucx_iprobe,
-        .pml_probe         = mca_pml_ucx_probe,
-        .pml_start         = mca_pml_ucx_start,
-        .pml_improbe       = mca_pml_ucx_improbe,
-        .pml_mprobe        = mca_pml_ucx_mprobe,
-        .pml_imrecv        = mca_pml_ucx_imrecv,
-        .pml_mrecv         = mca_pml_ucx_mrecv,
-        .pml_dump          = mca_pml_ucx_dump,
-        .pml_ft_event      = NULL,
-        .pml_max_contextid = 1ul << (PML_UCX_CONTEXT_BITS),
-        .pml_max_tag       = 1ul << (PML_UCX_TAG_BITS - 1)
+        .pml_add_procs      = mca_pml_ucx_add_procs,
+        .pml_del_procs      = mca_pml_ucx_del_procs,
+        .pml_enable         = mca_pml_ucx_enable,
+        .pml_progress       = NULL,
+        .pml_add_comm       = mca_pml_ucx_add_comm,
+        .pml_del_comm       = mca_pml_ucx_del_comm,
+        .pml_irecv_init     = mca_pml_ucx_irecv_init,
+        .pml_irecv          = mca_pml_ucx_irecv,
+        .pml_recv           = mca_pml_ucx_recv,
+        .pml_isend_init     = mca_pml_ucx_isend_init,
+        .pml_isend          = mca_pml_ucx_isend,
+        .pml_send           = mca_pml_ucx_send,
+        .pml_iprobe         = mca_pml_ucx_iprobe,
+        .pml_probe          = mca_pml_ucx_probe,
+        .pml_start          = mca_pml_ucx_start,
+        .pml_improbe        = mca_pml_ucx_improbe,
+        .pml_mprobe         = mca_pml_ucx_mprobe,
+        .pml_imrecv         = mca_pml_ucx_imrecv,
+        .pml_mrecv          = mca_pml_ucx_mrecv,
+        .pml_dump           = mca_pml_ucx_dump,
+        .pml_max_contextid  = (1ul << (PML_UCX_CONTEXT_BITS)) - 1,
+        .pml_max_tag        = (1ul << (PML_UCX_TAG_BITS - 1)) - 1,
+        .pml_flags          = 0, /* flags */
+        .pml_get_transports = mca_pml_ucx_get_transports
     },
     .ucp_context           = NULL,
     .ucp_worker            = NULL
@@ -109,7 +123,7 @@ static int mca_pml_ucx_send_worker_address_type(int addr_flags, int modex_scope)
     }
 
     PML_UCX_VERBOSE(2, "Pack %s worker address, size %ld",
-                    (modex_scope == OPAL_PMIX_LOCAL) ? "local" : "remote",
+                    (modex_scope == PMIX_LOCAL) ? "local" : "remote",
                     attrs.address_length);
 
     return OMPI_SUCCESS;
@@ -133,7 +147,7 @@ static int mca_pml_ucx_send_worker_address(void)
 
     PML_UCX_VERBOSE(2, "Pack worker address, size %ld", addrlen);
 
-    OPAL_MODEX_SEND(rc, OPAL_PMIX_GLOBAL,
+    OPAL_MODEX_SEND(rc, PMIX_GLOBAL,
                     &mca_pml_ucx_component.pmlm_version, (void*)address, addrlen);
 
     ucp_worker_release_address(ompi_pml_ucx.ucp_worker, address);
@@ -144,12 +158,12 @@ static int mca_pml_ucx_send_worker_address(void)
 #else
     /* Pack just network device addresses for remote node peers */
     status = mca_pml_ucx_send_worker_address_type(UCP_WORKER_ADDRESS_FLAG_NET_ONLY,
-                                                  OPAL_PMIX_REMOTE);
+                                                  PMIX_REMOTE);
     if (UCS_OK != status) {
         goto err;
     }
 
-    status = mca_pml_ucx_send_worker_address_type(0, OPAL_PMIX_LOCAL);
+    status = mca_pml_ucx_send_worker_address_type(0, PMIX_LOCAL);
     if (UCS_OK != status) {
         goto err;
     }
@@ -183,12 +197,29 @@ static int mca_pml_ucx_recv_worker_address(ompi_proc_t *proc,
 
 int mca_pml_ucx_open(void)
 {
+    unsigned major_version, minor_version, release_number;
     ucp_context_attr_t attr;
     ucp_params_t params;
     ucp_config_t *config;
     ucs_status_t status;
 
-    PML_UCX_VERBOSE(1, "mca_pml_ucx_open");
+    /* Check version */
+    ucp_get_version(&major_version, &minor_version, &release_number);
+    PML_UCX_VERBOSE(1, "mca_pml_ucx_open: UCX version %u.%u.%u",
+                    major_version, minor_version, release_number);
+
+    if ((major_version == 1) && (minor_version == 8)) {
+        /* disabled due to issue #8321 */
+        PML_UCX_VERBOSE(1, "UCX PML is disabled because the run-time UCX version "
+                           "is 1.8, which has a known catastrophic issue");
+        return OMPI_ERROR;
+    }
+
+    if ((major_version == 1) && (minor_version < 9)) {
+        /* show warning due to issue #8549 */
+        PML_UCX_WARN("UCX version %u.%u.%u is too old, please install 1.9.x or "
+                     "newer", major_version, minor_version, release_number);
+    }
 
     /* Read options */
     status = ucp_config_read("MPI", NULL, &config);
@@ -197,22 +228,32 @@ int mca_pml_ucx_open(void)
     }
 
     /* Initialize UCX context */
-    params.field_mask      = UCP_PARAM_FIELD_FEATURES |
-                             UCP_PARAM_FIELD_REQUEST_SIZE |
-                             UCP_PARAM_FIELD_REQUEST_INIT |
-                             UCP_PARAM_FIELD_REQUEST_CLEANUP |
-                             UCP_PARAM_FIELD_TAG_SENDER_MASK |
-                             UCP_PARAM_FIELD_MT_WORKERS_SHARED |
-                             UCP_PARAM_FIELD_ESTIMATED_NUM_EPS;
-    params.features        = UCP_FEATURE_TAG;
-    params.request_size    = sizeof(ompi_request_t);
-    params.request_init    = mca_pml_ucx_request_init;
-    params.request_cleanup = mca_pml_ucx_request_cleanup;
-    params.tag_sender_mask = PML_UCX_SPECIFIC_SOURCE_MASK;
+    params.field_mask        = UCP_PARAM_FIELD_FEATURES |
+                               UCP_PARAM_FIELD_REQUEST_SIZE |
+                               UCP_PARAM_FIELD_REQUEST_INIT |
+                               UCP_PARAM_FIELD_REQUEST_CLEANUP |
+                               UCP_PARAM_FIELD_TAG_SENDER_MASK |
+                               UCP_PARAM_FIELD_MT_WORKERS_SHARED |
+                               UCP_PARAM_FIELD_ESTIMATED_NUM_EPS;
+    params.features          = UCP_FEATURE_TAG;
+    params.request_size      = sizeof(ompi_request_t);
+    params.request_init      = mca_pml_ucx_request_init;
+    params.request_cleanup   = mca_pml_ucx_request_cleanup;
+    params.tag_sender_mask   = PML_UCX_SPECIFIC_SOURCE_MASK;
     params.mt_workers_shared = 0; /* we do not need mt support for context
                                      since it will be protected by worker */
     params.estimated_num_eps = ompi_proc_world_size();
 
+#if HAVE_DECL_UCP_PARAM_FIELD_ESTIMATED_NUM_PPN
+    params.estimated_num_ppn = opal_process_info.num_local_peers + 1;
+    params.field_mask       |= UCP_PARAM_FIELD_ESTIMATED_NUM_PPN;
+#endif
+
+#if HAVE_DECL_UCP_PARAM_FIELD_NODE_LOCAL_ID
+    params.node_local_id = opal_process_info.my_local_rank;
+    params.field_mask   |= UCP_PARAM_FIELD_NODE_LOCAL_ID;
+#endif
+
     status = ucp_init(&params, config, &ompi_pml_ucx.ucp_context);
     ucp_config_release(config);
 
@@ -222,6 +263,9 @@ int mca_pml_ucx_open(void)
 
     /* Query UCX attributes */
     attr.field_mask        = UCP_ATTR_FIELD_REQUEST_SIZE;
+#if HAVE_UCP_ATTR_MEMORY_TYPES
+    attr.field_mask       |= UCP_ATTR_FIELD_MEMORY_TYPES;
+#endif
     status = ucp_context_query(ompi_pml_ucx.ucp_context, &attr);
     if (UCS_OK != status) {
         ucp_cleanup(ompi_pml_ucx.ucp_context);
@@ -229,7 +273,7 @@ int mca_pml_ucx_open(void)
         return OMPI_ERROR;
     }
 
-    ompi_pml_ucx.request_size = attr.request_size;
+    ompi_pml_ucx.request_size     = attr.request_size;
 
     return OMPI_SUCCESS;
 }
@@ -245,7 +289,7 @@ int mca_pml_ucx_close(void)
     return OMPI_SUCCESS;
 }
 
-int mca_pml_ucx_init(void)
+int mca_pml_ucx_init(int enable_mpi_threads)
 {
     ucp_worker_params_t params;
     ucp_worker_attr_t attr;
@@ -254,15 +298,21 @@ int mca_pml_ucx_init(void)
 
     PML_UCX_VERBOSE(1, "mca_pml_ucx_init");
 
-    /* TODO check MPI thread mode */
     params.field_mask  = UCP_WORKER_PARAM_FIELD_THREAD_MODE;
-    params.thread_mode = UCS_THREAD_MODE_SINGLE;
-    if (ompi_mpi_thread_multiple) {
+    if (enable_mpi_threads) {
         params.thread_mode = UCS_THREAD_MODE_MULTI;
     } else {
-        params.thread_mode = UCS_THREAD_MODE_SINGLE;
+        params.thread_mode =
+            opal_common_ucx_thread_mode(ompi_mpi_thread_provided);
     }
 
+#if HAVE_DECL_UCP_WORKER_FLAG_IGNORE_REQUEST_LEAK
+    if (!ompi_pml_ucx.request_leak_check) {
+        params.field_mask |= UCP_WORKER_PARAM_FIELD_FLAGS;
+        params.flags      |= UCP_WORKER_FLAG_IGNORE_REQUEST_LEAK;
+    }
+#endif
+
     status = ucp_worker_create(ompi_pml_ucx.ucp_context, &params,
                                &ompi_pml_ucx.ucp_worker);
     if (UCS_OK != status) {
@@ -279,10 +329,11 @@ int mca_pml_ucx_init(void)
         goto err_destroy_worker;
     }
 
-    if (ompi_mpi_thread_multiple && (attr.thread_mode != UCS_THREAD_MODE_MULTI)) {
+    if (enable_mpi_threads && (attr.thread_mode != UCS_THREAD_MODE_MULTI)) {
         /* UCX does not support multithreading, disqualify current PML for now */
         /* TODO: we should let OMPI to fallback to THREAD_SINGLE mode */
-        PML_UCX_ERROR("UCP worker does not support MPI_THREAD_MULTIPLE");
+        PML_UCX_WARN("UCP worker does not support MPI_THREAD_MULTIPLE. "
+                     "PML UCX could not be selected");
         rc = OMPI_ERR_NOT_SUPPORTED;
         goto err_destroy_worker;
     }
@@ -304,19 +355,22 @@ int mca_pml_ucx_init(void)
     /* Create a completed request to be returned from isend */
     OBJ_CONSTRUCT(&ompi_pml_ucx.completed_send_req, ompi_request_t);
     mca_pml_ucx_completed_request_init(&ompi_pml_ucx.completed_send_req);
+#if MPI_VERSION >= 4
+    ompi_pml_ucx.completed_send_req.req_cancel = mca_pml_cancel_send_callback;
+#endif
 
     opal_progress_register(mca_pml_ucx_progress);
 
     PML_UCX_VERBOSE(2, "created ucp context %p, worker %p",
                     (void *)ompi_pml_ucx.ucp_context,
                     (void *)ompi_pml_ucx.ucp_worker);
-    return rc;
+    return OMPI_SUCCESS;
 
 err_destroy_worker:
     ucp_worker_destroy(ompi_pml_ucx.ucp_worker);
-    ompi_pml_ucx.ucp_worker = NULL;
 err:
-    return OMPI_ERROR;
+    ompi_pml_ucx.ucp_worker = NULL;
+    return rc;
 }
 
 int mca_pml_ucx_cleanup(void)
@@ -345,7 +399,7 @@ int mca_pml_ucx_cleanup(void)
     OBJ_DESTRUCT(&ompi_pml_ucx.convs);
     OBJ_DESTRUCT(&ompi_pml_ucx.persistent_reqs);
 
-    if (ompi_pml_ucx.ucp_worker) {
+    if (ompi_pml_ucx.ucp_worker != NULL) {
         ucp_worker_destroy(ompi_pml_ucx.ucp_worker);
         ompi_pml_ucx.ucp_worker = NULL;
     }
@@ -355,13 +409,18 @@ int mca_pml_ucx_cleanup(void)
 
 static ucp_ep_h mca_pml_ucx_add_proc_common(ompi_proc_t *proc)
 {
+    size_t addrlen = 0;
     ucp_ep_params_t ep_params;
     ucp_address_t *address;
     ucs_status_t status;
-    size_t addrlen;
     ucp_ep_h ep;
     int ret;
 
+    /* Do not add a new endpoint if we already created one */
+    if (NULL != proc->proc_endpoints[OMPI_PROC_ENDPOINT_TAG_PML]) {
+        return proc->proc_endpoints[OMPI_PROC_ENDPOINT_TAG_PML];
+    }
+
     ret = mca_pml_ucx_recv_worker_address(proc, &address, &addrlen);
     if (ret < 0) {
         return NULL;
@@ -385,34 +444,11 @@ static ucp_ep_h mca_pml_ucx_add_proc_common(ompi_proc_t *proc)
     return ep;
 }
 
-static ucp_ep_h mca_pml_ucx_add_proc(ompi_communicator_t *comm, int dst)
-{
-    ompi_proc_t *proc0     = ompi_comm_peer_lookup(comm, 0);
-    ompi_proc_t *proc_peer = ompi_comm_peer_lookup(comm, dst);
-    int ret;
-
-    /* Note, mca_pml_base_pml_check_selected, doesn't use 3rd argument */
-    if (OMPI_SUCCESS != (ret = mca_pml_base_pml_check_selected("ucx",
-                                                               &proc0,
-                                                               dst))) {
-        return NULL;
-    }
-
-    return mca_pml_ucx_add_proc_common(proc_peer);
-}
-
 int mca_pml_ucx_add_procs(struct ompi_proc_t **procs, size_t nprocs)
 {
     ompi_proc_t *proc;
     ucp_ep_h ep;
     size_t i;
-    int ret;
-
-    if (OMPI_SUCCESS != (ret = mca_pml_base_pml_check_selected("ucx",
-                                                               procs,
-                                                               nprocs))) {
-        return ret;
-    }
 
     for (i = 0; i < nprocs; ++i) {
         proc = procs[(i + OMPI_PROC_MY_NAME->vpid) % nprocs];
@@ -426,28 +462,77 @@ int mca_pml_ucx_add_procs(struct ompi_proc_t **procs, size_t nprocs)
     return OMPI_SUCCESS;
 }
 
+mca_pml_transports_t *mca_pml_ucx_get_transports(struct ompi_communicator_t *comm,
+                                                 int rank)
+{
+#if HAVE_DECL_UCP_EP_ATTR_FIELD_TRANSPORTS
+    mca_pml_transports_t *transports;
+    unsigned i;
+    ucs_status_t status;
+    ucp_ep_attr_t ep_attrs;
+    ompi_proc_t *proc = ompi_comm_peer_lookup(comm, rank);
+
+    if (NULL == proc) {
+        return NULL;
+    }
+    if (NULL == proc->proc_endpoints[OMPI_PROC_ENDPOINT_TAG_PML]) {
+        return NULL;
+    }
+
+    ep_attrs.field_mask = UCP_EP_ATTR_FIELD_TRANSPORTS;
+    ep_attrs.transports.num_entries = PML_UCX_MAX_TRANSPORT_ENTRIES;
+    ep_attrs.transports.entry_size = sizeof(ucp_transport_entry_t);
+    ep_attrs.transports.entries = (ucp_transport_entry_t *) malloc(
+             sizeof(ucp_transport_entry_t) * ep_attrs.transports.num_entries);
+    if (NULL == ep_attrs.transports.entries) {
+        return NULL;
+    }
+
+    status = ucp_ep_query(proc->proc_endpoints[OMPI_PROC_ENDPOINT_TAG_PML],
+                          &ep_attrs);
+    if (UCS_OK != status) {
+        return NULL;
+    }
+
+    transports = (mca_pml_transports_t *) malloc(sizeof(mca_pml_transports_t));
+    if (NULL == transports) {
+        return NULL;
+    }
+
+    transports->entries = (mca_pml_transport_entry_t *) malloc(
+        ep_attrs.transports.num_entries * sizeof(mca_pml_transport_entry_t));
+    if (NULL == transports->entries) {
+        free(transports);
+        return NULL;
+    }
+
+    for (i = 0; i < ep_attrs.transports.num_entries; i++) {
+        transports->entries[i].transport_name =
+                ep_attrs.transports.entries[i].transport_name;
+        transports->entries[i].device_name = 
+                ep_attrs.transports.entries[i].device_name;
+    }
+    transports->count = ep_attrs.transports.num_entries;
+
+    free(ep_attrs.transports.entries);
+    return transports;
+#else
+    return NULL;
+#endif
+}
+
+__opal_attribute_always_inline__
 static inline ucp_ep_h mca_pml_ucx_get_ep(ompi_communicator_t *comm, int rank)
 {
+    ompi_proc_t *proc_peer = ompi_comm_peer_lookup(comm, rank);
     ucp_ep_h ep;
 
-    ep = ompi_comm_peer_lookup(comm, rank)->proc_endpoints[OMPI_PROC_ENDPOINT_TAG_PML];
-    if (OPAL_LIKELY(ep != NULL)) {
+    ep = proc_peer->proc_endpoints[OMPI_PROC_ENDPOINT_TAG_PML];
+    if (OPAL_LIKELY(NULL != ep)) {
         return ep;
     }
 
-    ep = mca_pml_ucx_add_proc(comm, rank);
-    if (OPAL_LIKELY(ep != NULL)) {
-        return ep;
-    }
-
-    if (rank >= ompi_comm_size(comm)) {
-        PML_UCX_ERROR("Rank number (%d) is larger than communicator size (%d)",
-                      rank, ompi_comm_size(comm));
-    } else {
-        PML_UCX_ERROR("Failed to resolve UCX endpoint for rank %d", rank);
-    }
-
-    return NULL;
+    return mca_pml_ucx_add_proc_common(proc_peer);
 }
 
 int mca_pml_ucx_del_procs(struct ompi_proc_t **procs, size_t nprocs)
@@ -485,8 +570,7 @@ int mca_pml_ucx_enable(bool enable)
     int ret;
 
     /* Create a key for adding custom attributes to datatypes */
-    copy_fn.attr_datatype_copy_fn  =
-                    (MPI_Type_internal_copy_attr_function*)MPI_TYPE_NULL_COPY_FN;
+    copy_fn.attr_datatype_copy_fn  = MPI_TYPE_NULL_COPY_FN;
     del_fn.attr_datatype_delete_fn = mca_pml_ucx_datatype_attr_del_fn;
     ret = ompi_attr_create_keyval(TYPE_ATTR, copy_fn, del_fn,
                                   &ompi_pml_ucx.datatype_attr_keyval, NULL, 0,
@@ -507,8 +591,7 @@ int mca_pml_ucx_enable(bool enable)
 
 int mca_pml_ucx_progress(void)
 {
-    ucp_worker_progress(ompi_pml_ucx.ucp_worker);
-    return OMPI_SUCCESS;
+    return ucp_worker_progress(ompi_pml_ucx.ucp_worker);
 }
 
 int mca_pml_ucx_add_comm(struct ompi_communicator_t* comm)
@@ -540,7 +623,9 @@ int mca_pml_ucx_irecv_init(void *buf, size_t count, ompi_datatype_t *datatype,
     req->flags                    = 0;
     req->buffer                   = buf;
     req->count                    = count;
-    req->datatype.datatype        = mca_pml_ucx_get_datatype(datatype);
+    req->ompi_datatype            = datatype;
+    req->datatype                 = mca_pml_ucx_get_datatype(datatype);
+    OMPI_DATATYPE_RETAIN(datatype);
 
     PML_UCX_MAKE_RECV_TAG(req->tag, req->recv.tag_mask, tag, src, comm);
 
@@ -552,6 +637,11 @@ int mca_pml_ucx_irecv(void *buf, size_t count, ompi_datatype_t *datatype,
                       int src, int tag, struct ompi_communicator_t* comm,
                       struct ompi_request_t **request)
 {
+#if HAVE_DECL_UCP_TAG_RECV_NBX
+    pml_ucx_datatype_t *op_data = mca_pml_ucx_get_op_data(datatype);
+    ucp_request_param_t *param  = &op_data->op_param.irecv;
+#endif
+
     ucp_tag_t ucp_tag, ucp_tag_mask;
     ompi_request_t *req;
 
@@ -559,10 +649,16 @@ int mca_pml_ucx_irecv(void *buf, size_t count, ompi_datatype_t *datatype,
                        (void*)request);
 
     PML_UCX_MAKE_RECV_TAG(ucp_tag, ucp_tag_mask, tag, src, comm);
+#if HAVE_DECL_UCP_TAG_RECV_NBX
+    req = (ompi_request_t*)ucp_tag_recv_nbx(ompi_pml_ucx.ucp_worker, buf,
+                                            mca_pml_ucx_get_data_size(op_data, count),
+                                            ucp_tag, ucp_tag_mask, param);
+#else
     req = (ompi_request_t*)ucp_tag_recv_nb(ompi_pml_ucx.ucp_worker, buf, count,
                                            mca_pml_ucx_get_datatype(datatype),
                                            ucp_tag, ucp_tag_mask,
                                            mca_pml_ucx_recv_completion);
+#endif
     if (UCS_PTR_IS_ERR(req)) {
         PML_UCX_ERROR("ucx recv failed: %s", ucs_status_string(UCS_PTR_STATUS(req)));
         return OMPI_ERROR;
@@ -578,29 +674,52 @@ int mca_pml_ucx_recv(void *buf, size_t count, ompi_datatype_t *datatype, int src
                      int tag, struct ompi_communicator_t* comm,
                      ompi_status_public_t* mpi_status)
 {
+    /* coverity[bad_alloc_arithmetic] */
+    void *req = PML_UCX_REQ_ALLOCA();
+#if HAVE_DECL_UCP_TAG_RECV_NBX
+    pml_ucx_datatype_t *op_data     = mca_pml_ucx_get_op_data(datatype);
+    ucp_request_param_t *recv_param = &op_data->op_param.recv;
+    ucp_request_param_t param;
+
+    param.op_attr_mask = UCP_OP_ATTR_FIELD_REQUEST |
+                         (recv_param->op_attr_mask & UCP_OP_ATTR_FIELD_DATATYPE);
+    param.datatype     = recv_param->datatype;
+    param.request      = req;
+#endif
     ucp_tag_t ucp_tag, ucp_tag_mask;
     ucp_tag_recv_info_t info;
     ucs_status_t status;
-    void *req;
+    int result;
 
     PML_UCX_TRACE_RECV("%s", buf, count, datatype, src, tag, comm, "recv");
 
-    /* coverity[bad_alloc_arithmetic] */
     PML_UCX_MAKE_RECV_TAG(ucp_tag, ucp_tag_mask, tag, src, comm);
-    req = PML_UCX_REQ_ALLOCA();
-    status = ucp_tag_recv_nbr(ompi_pml_ucx.ucp_worker, buf, count,
-                              mca_pml_ucx_get_datatype(datatype),
-                              ucp_tag, ucp_tag_mask, req);
-
+#if HAVE_DECL_UCP_TAG_RECV_NBX
+    ucp_tag_recv_nbx(ompi_pml_ucx.ucp_worker, buf,
+                     mca_pml_ucx_get_data_size(op_data, count),
+                     ucp_tag, ucp_tag_mask, &param);
+#else
+    ucp_tag_recv_nbr(ompi_pml_ucx.ucp_worker, buf, count,
+                     mca_pml_ucx_get_datatype(datatype),
+                     ucp_tag, ucp_tag_mask, req);
+#endif
     MCA_COMMON_UCX_PROGRESS_LOOP(ompi_pml_ucx.ucp_worker) {
         status = ucp_request_test(req, &info);
         if (status != UCS_INPROGRESS) {
-            mca_pml_ucx_set_recv_status_safe(mpi_status, status, &info);
-            return OMPI_SUCCESS;
+            result = mca_pml_ucx_set_recv_status_public(mpi_status, status, &info);
+
+#if SPC_ENABLE == 1
+            size_t dt_size;
+            ompi_datatype_type_size(datatype, &dt_size);
+            SPC_USER_OR_MPI(tag, dt_size*count,
+                            OMPI_SPC_BYTES_RECEIVED_USER, OMPI_SPC_BYTES_RECEIVED_MPI);
+#endif
+            return result;
         }
     }
 }
 
+__opal_attribute_always_inline__
 static inline const char *mca_pml_ucx_send_mode_name(mca_pml_base_send_mode_t mode)
 {
     switch (mode) {
@@ -650,12 +769,13 @@ int mca_pml_ucx_isend_init(const void *buf, size_t count, ompi_datatype_t *datat
     req->tag                      = PML_UCX_MAKE_SEND_TAG(tag, comm);
     req->send.mode                = mode;
     req->send.ep                  = ep;
+    req->ompi_datatype            = datatype;
+    OMPI_DATATYPE_RETAIN(datatype);
 
     if (MCA_PML_BASE_SEND_BUFFERED == mode) {
-        req->datatype.ompi_datatype = datatype;
-        OBJ_RETAIN(datatype);
+        req->datatype = (ucp_datatype_t)NULL;
     } else {
-        req->datatype.datatype = mca_pml_ucx_get_datatype(datatype);
+        req->datatype = mca_pml_ucx_get_datatype(datatype);
     }
 
     *request = &req->ompi;
@@ -663,7 +783,7 @@ int mca_pml_ucx_isend_init(const void *buf, size_t count, ompi_datatype_t *datat
 }
 
 static ucs_status_ptr_t
-mca_pml_ucx_bsend(ucp_ep_h ep, const void *buf, size_t count, 
+mca_pml_ucx_bsend(ucp_ep_h ep, const void *buf, size_t count,
                   ompi_datatype_t *datatype, uint64_t pml_tag)
 {
     ompi_request_t *req;
@@ -686,7 +806,7 @@ mca_pml_ucx_bsend(ucp_ep_h ep, const void *buf, size_t count,
         PML_UCX_ERROR("bsend: failed to allocate buffer");
         return UCS_STATUS_PTR(OMPI_ERROR);
     }
-    
+
     iov_count    = 1;
     iov.iov_base = packed_data;
     iov.iov_len  = packed_length;
@@ -722,6 +842,7 @@ mca_pml_ucx_bsend(ucp_ep_h ep, const void *buf, size_t count,
     return NULL;
 }
 
+__opal_attribute_always_inline__
 static inline ucs_status_ptr_t mca_pml_ucx_common_send(ucp_ep_h ep, const void *buf,
                                                        size_t count,
                                                        ompi_datatype_t *datatype,
@@ -739,6 +860,32 @@ static inline ucs_status_ptr_t mca_pml_ucx_common_send(ucp_ep_h ep, const void *
     }
 }
 
+#if HAVE_DECL_UCP_TAG_SEND_NBX
+__opal_attribute_always_inline__
+static inline ucs_status_ptr_t
+mca_pml_ucx_common_send_nbx(ucp_ep_h ep, const void *buf,
+                            size_t count,
+                            ompi_datatype_t *datatype,
+                            ucp_tag_t tag,
+                            mca_pml_base_send_mode_t mode,
+                            ucp_request_param_t *param)
+{
+    pml_ucx_datatype_t *op_data = mca_pml_ucx_get_op_data(datatype);
+
+    if (OPAL_UNLIKELY(MCA_PML_BASE_SEND_BUFFERED == mode)) {
+        return mca_pml_ucx_bsend(ep, buf, count, datatype, tag);
+    } else if (OPAL_UNLIKELY(MCA_PML_BASE_SEND_SYNCHRONOUS == mode)) {
+        return ucp_tag_send_sync_nb(ep, buf, count,
+                                    mca_pml_ucx_get_datatype(datatype), tag,
+                                    (ucp_send_callback_t)param->cb.send);
+    } else {
+        return ucp_tag_send_nbx(ep, buf,
+                                mca_pml_ucx_get_data_size(op_data, count),
+                                tag, param);
+    }
+}
+#endif
+
 int mca_pml_ucx_isend(const void *buf, size_t count, ompi_datatype_t *datatype,
                       int dst, int tag, mca_pml_base_send_mode_t mode,
                       struct ompi_communicator_t* comm,
@@ -747,8 +894,8 @@ int mca_pml_ucx_isend(const void *buf, size_t count, ompi_datatype_t *datatype,
     ompi_request_t *req;
     ucp_ep_h ep;
 
-    PML_UCX_TRACE_SEND("i%ssend request *%p", 
-                       buf, count, datatype, dst, tag, mode, comm, 
+    PML_UCX_TRACE_SEND("i%ssend request *%p",
+                       buf, count, datatype, dst, tag, mode, comm,
                        mode == MCA_PML_BASE_SEND_BUFFERED ? "b" : "",
                        (void*)request)
 
@@ -757,10 +904,23 @@ int mca_pml_ucx_isend(const void *buf, size_t count, ompi_datatype_t *datatype,
         return OMPI_ERROR;
     }
 
+#if HAVE_DECL_UCP_TAG_SEND_NBX
+    req = (ompi_request_t*)mca_pml_ucx_common_send_nbx(ep, buf, count, datatype,
+                                                       PML_UCX_MAKE_SEND_TAG(tag, comm), mode,
+                                                       &mca_pml_ucx_get_op_data(datatype)->op_param.isend);
+#else
     req = (ompi_request_t*)mca_pml_ucx_common_send(ep, buf, count, datatype,
                                                    mca_pml_ucx_get_datatype(datatype),
                                                    PML_UCX_MAKE_SEND_TAG(tag, comm), mode,
                                                    mca_pml_ucx_send_completion);
+#endif
+
+#if SPC_ENABLE == 1
+    size_t dt_size;
+    ompi_datatype_type_size(datatype, &dt_size);
+    SPC_USER_OR_MPI(tag, dt_size*count,
+                    OMPI_SPC_BYTES_SENT_USER, OMPI_SPC_BYTES_SENT_MPI);
+#endif
 
     if (req == NULL) {
         PML_UCX_VERBOSE(8, "returning completed request");
@@ -769,6 +929,11 @@ int mca_pml_ucx_isend(const void *buf, size_t count, ompi_datatype_t *datatype,
     } else if (!UCS_PTR_IS_ERR(req)) {
         PML_UCX_VERBOSE(8, "got request %p", (void*)req);
         req->req_mpi_object.comm = comm;
+#if MPI_VERSION >= 4
+        if (OPAL_LIKELY(mca_pml_ucx_request_cancel == req->req_cancel)) {
+            req->req_cancel      = mca_pml_ucx_request_cancel_send;
+        }
+#endif
         *request                 = req;
         return OMPI_SUCCESS;
     } else {
@@ -780,19 +945,19 @@ int mca_pml_ucx_isend(const void *buf, size_t count, ompi_datatype_t *datatype,
 static inline __opal_attribute_always_inline__ int
 mca_pml_ucx_send_nb(ucp_ep_h ep, const void *buf, size_t count,
                     ompi_datatype_t *datatype, ucp_datatype_t ucx_datatype,
-                    ucp_tag_t tag, mca_pml_base_send_mode_t mode,
-                    ucp_send_callback_t cb)
+                    ucp_tag_t tag, mca_pml_base_send_mode_t mode)
 {
     ompi_request_t *req;
 
     req = (ompi_request_t*)mca_pml_ucx_common_send(ep, buf, count, datatype,
                                                    mca_pml_ucx_get_datatype(datatype),
-                                                   tag, mode, cb);
+                                                   tag, mode,
+                                                   mca_pml_ucx_send_completion_empty);
     if (OPAL_LIKELY(req == NULL)) {
         return OMPI_SUCCESS;
     } else if (!UCS_PTR_IS_ERR(req)) {
         PML_UCX_VERBOSE(8, "got request %p", (void*)req);
-        MCA_COMMON_UCX_WAIT_LOOP(req, ompi_pml_ucx.ucp_worker, "ucx send", ompi_request_free(&req));
+        MCA_COMMON_UCX_WAIT_LOOP(req, ompi_pml_ucx.ucp_worker, "ucx send", ucp_request_free(req));
     } else {
         PML_UCX_ERROR("ucx send failed: %s", ucs_status_string(UCS_PTR_STATUS(req)));
         return OMPI_ERROR;
@@ -802,20 +967,40 @@ mca_pml_ucx_send_nb(ucp_ep_h ep, const void *buf, size_t count,
 #if HAVE_DECL_UCP_TAG_SEND_NBR
 static inline __opal_attribute_always_inline__ int
 mca_pml_ucx_send_nbr(ucp_ep_h ep, const void *buf, size_t count,
-                     ucp_datatype_t ucx_datatype, ucp_tag_t tag)
-
+                     ompi_datatype_t *datatype, ucp_tag_t tag)
 {
-    ucs_status_ptr_t req;
-    ucs_status_t status;
-
     /* coverity[bad_alloc_arithmetic] */
-    req    = PML_UCX_REQ_ALLOCA();
-    status = ucp_tag_send_nbr(ep, buf, count, ucx_datatype, tag, req);
+    ucs_status_ptr_t req = PML_UCX_REQ_ALLOCA();
+#if HAVE_DECL_UCP_TAG_SEND_NBX
+    pml_ucx_datatype_t *op_data = mca_pml_ucx_get_op_data(datatype);
+    ucp_request_param_t param   = {
+        .op_attr_mask = UCP_OP_ATTR_FIELD_REQUEST |
+                        (op_data->op_param.send.op_attr_mask & UCP_OP_ATTR_FIELD_DATATYPE) |
+                        UCP_OP_ATTR_FLAG_FAST_CMPL,
+        .datatype     = op_data->op_param.send.datatype,
+        .request      = req
+    };
+
+    req = ucp_tag_send_nbx(ep, buf,
+                           mca_pml_ucx_get_data_size(op_data, count),
+                           tag, &param);
+    if (OPAL_LIKELY(req == UCS_OK)) {
+        return OMPI_SUCCESS;
+    } else if (UCS_PTR_IS_ERR(req)) {
+        PML_UCX_ERROR("%s failed: %d, %s", __func__, UCS_PTR_STATUS(req),
+                      ucs_status_string(UCS_PTR_STATUS(req)));
+        return OPAL_ERROR;
+    }
+#else
+    ucs_status_t status;
+    status = ucp_tag_send_nbr(ep, buf, count,
+                              mca_pml_ucx_get_datatype(datatype), tag, req);
     if (OPAL_LIKELY(status == UCS_OK)) {
         return OMPI_SUCCESS;
     }
+#endif
 
-    MCA_COMMON_UCX_WAIT_LOOP(req, ompi_pml_ucx.ucp_worker, "ucx send", (void)0);
+    MCA_COMMON_UCX_WAIT_LOOP(req, ompi_pml_ucx.ucp_worker, "ucx send nbr", (void)0);
 }
 #endif
 
@@ -833,19 +1018,24 @@ int mca_pml_ucx_send(const void *buf, size_t count, ompi_datatype_t *datatype, i
         return OMPI_ERROR;
     }
 
+#if SPC_ENABLE == 1
+    size_t dt_size;
+    ompi_datatype_type_size(datatype, &dt_size);
+    SPC_USER_OR_MPI(tag, dt_size*count,
+                    OMPI_SPC_BYTES_SENT_USER, OMPI_SPC_BYTES_SENT_MPI);
+#endif
+
 #if HAVE_DECL_UCP_TAG_SEND_NBR
     if (OPAL_LIKELY((MCA_PML_BASE_SEND_BUFFERED != mode) &&
                     (MCA_PML_BASE_SEND_SYNCHRONOUS != mode))) {
-        return mca_pml_ucx_send_nbr(ep, buf, count,
-                                    mca_pml_ucx_get_datatype(datatype),
+        return mca_pml_ucx_send_nbr(ep, buf, count, datatype,
                                     PML_UCX_MAKE_SEND_TAG(tag, comm));
     }
 #endif
 
     return mca_pml_ucx_send_nb(ep, buf, count, datatype,
                                mca_pml_ucx_get_datatype(datatype),
-                               PML_UCX_MAKE_SEND_TAG(tag, comm), mode,
-                               mca_pml_ucx_send_completion);
+                               PML_UCX_MAKE_SEND_TAG(tag, comm), mode);
 }
 
 int mca_pml_ucx_iprobe(int src, int tag, struct ompi_communicator_t* comm,
@@ -864,10 +1054,10 @@ int mca_pml_ucx_iprobe(int src, int tag, struct ompi_communicator_t* comm,
                                0, &info);
     if (ucp_msg != NULL) {
         *matched = 1;
-        mca_pml_ucx_set_recv_status_safe(mpi_status, UCS_OK, &info);
+        mca_pml_ucx_set_recv_status_public(mpi_status, UCS_OK, &info);
     } else  {
         (++progress_count % opal_common_ucx.progress_iterations) ?
-            (void)ucp_worker_progress(ompi_pml_ucx.ucp_worker) : opal_progress();
+            (int)ucp_worker_progress(ompi_pml_ucx.ucp_worker) : opal_progress();
         *matched = 0;
     }
     return OMPI_SUCCESS;
@@ -888,7 +1078,7 @@ int mca_pml_ucx_probe(int src, int tag, struct ompi_communicator_t* comm,
         ucp_msg = ucp_tag_probe_nb(ompi_pml_ucx.ucp_worker, ucp_tag,
                                    ucp_tag_mask, 0, &info);
         if (ucp_msg != NULL) {
-            mca_pml_ucx_set_recv_status_safe(mpi_status, UCS_OK, &info);
+            mca_pml_ucx_set_recv_status_public(mpi_status, UCS_OK, &info);
             return OMPI_SUCCESS;
         }
     }
@@ -913,10 +1103,10 @@ int mca_pml_ucx_improbe(int src, int tag, struct ompi_communicator_t* comm,
         PML_UCX_MESSAGE_NEW(comm, ucp_msg, &info, message);
         PML_UCX_VERBOSE(8, "got message %p (%p)", (void*)*message, (void*)ucp_msg);
         *matched         = 1;
-        mca_pml_ucx_set_recv_status_safe(mpi_status, UCS_OK, &info);
+        mca_pml_ucx_set_recv_status_public(mpi_status, UCS_OK, &info);
     } else  {
         (++progress_count % opal_common_ucx.progress_iterations) ?
-            (void)ucp_worker_progress(ompi_pml_ucx.ucp_worker) : opal_progress();
+            (int)ucp_worker_progress(ompi_pml_ucx.ucp_worker) : opal_progress();
         *matched = 0;
     }
     return OMPI_SUCCESS;
@@ -939,7 +1129,7 @@ int mca_pml_ucx_mprobe(int src, int tag, struct ompi_communicator_t* comm,
         if (ucp_msg != NULL) {
             PML_UCX_MESSAGE_NEW(comm, ucp_msg, &info, message);
             PML_UCX_VERBOSE(8, "got message %p (%p)", (void*)*message, (void*)ucp_msg);
-            mca_pml_ucx_set_recv_status_safe(mpi_status, UCS_OK, &info);
+            mca_pml_ucx_set_recv_status_public(mpi_status, UCS_OK, &info);
             return OMPI_SUCCESS;
         }
     }
@@ -963,6 +1153,7 @@ int mca_pml_ucx_imrecv(void *buf, size_t count, ompi_datatype_t *datatype,
     }
 
     PML_UCX_VERBOSE(8, "got request %p", (void*)req);
+    req->req_mpi_object.comm = (*message)->comm;
     PML_UCX_MESSAGE_RELEASE(message);
     *request = req;
     return OMPI_SUCCESS;
@@ -985,10 +1176,10 @@ int mca_pml_ucx_mrecv(void *buf, size_t count, ompi_datatype_t *datatype,
         return OMPI_ERROR;
     }
 
+    req->req_mpi_object.comm = (*message)->comm;
     PML_UCX_MESSAGE_RELEASE(message);
 
-    ompi_request_wait(&req, status);
-    return OMPI_SUCCESS;
+    return ompi_request_wait(&req, status);
 }
 
 int mca_pml_ucx_start(size_t count, ompi_request_t** requests)
@@ -1013,8 +1204,8 @@ int mca_pml_ucx_start(size_t count, ompi_request_t** requests)
             tmp_req = (ompi_request_t*)mca_pml_ucx_common_send(preq->send.ep,
                                                                preq->buffer,
                                                                preq->count,
-                                                               preq->datatype.ompi_datatype,
-                                                               preq->datatype.datatype,
+                                                               preq->ompi_datatype,
+                                                               preq->datatype,
                                                                preq->tag,
                                                                preq->send.mode,
                                                                mca_pml_ucx_psend_completion);
@@ -1022,7 +1213,7 @@ int mca_pml_ucx_start(size_t count, ompi_request_t** requests)
             PML_UCX_VERBOSE(8, "start recv request %p", (void*)preq);
             tmp_req = (ompi_request_t*)ucp_tag_recv_nb(ompi_pml_ucx.ucp_worker,
                                                        preq->buffer, preq->count,
-                                                       preq->datatype.datatype,
+                                                       preq->datatype,
                                                        preq->tag,
                                                        preq->recv.tag_mask,
                                                        mca_pml_ucx_precv_completion);
