@@ -1,3 +1,4 @@
+/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil -*- */
 /*
  * Copyright (c) 2004-2007 The Trustees of Indiana University and Indiana
  *                         University Research and Technology
@@ -5,14 +6,16 @@
  * Copyright (c) 2004-2006 The University of Tennessee and The University
  *                         of Tennessee Research Foundation.  All rights
  *                         reserved.
- * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart, 
+ * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart,
  *                         University of Stuttgart.  All rights reserved.
  * Copyright (c) 2004-2005 The Regents of the University of California.
  *                         All rights reserved.
+ * Copyright (c) 2018      Triad National Security, LLC. All rights
+ *                         reserved.
  * $COPYRIGHT$
- * 
+ *
  * Additional copyrights may follow
- * 
+ *
  * $HEADER$
  */
 
@@ -36,20 +39,19 @@
  */
 
 #ifndef OPAL_MEMORY_MEMORY_H
-#define OPAl_MEMORY_MEMORY_H
+#define OPAL_MEMORY_MEMORY_H
 
 #include "opal_config.h"
 #include "memory_internal.h"
 
 BEGIN_C_DECLS
 
-
 /**
  * Initialize the memory hooks subsystem
  *
  * Initialize the memory hooks subsystem.  This is generally called
  * during opal_init() and should be called before any other function
- * in the interface is called.  
+ * in the interface is called.
  *
  * \note Note that some back-end functionality is activated pre-main,
  * so not calling this function does not prevent the memory hooks from
@@ -59,25 +61,6 @@ BEGIN_C_DECLS
  */
 OPAL_DECLSPEC int opal_mem_hooks_init(void);
 
-
-/**
- * Finalize the memory hooks subsystem
- *
- * Finalize the memory hooks subsystem.  This is generally called
- * during opal_finalize() and no other memory hooks functions should
- * be called after this function is called.  opal_mem_hooks_finalize()
- * will automatically deregister any callbacks that have not already
- * been deregistered.  In a multi-threaded application, it is possible
- * that one thread will have a memory hook callback while the other
- * thread is in opal_mem_hooks_finalize(), however, no threads will
- * receive a callback once the calling thread has exited
- * opal_mem_hooks_finalize().
- *
- * @retval OPAL_SUCCESS Shutdown completed successfully
- */
-OPAL_DECLSPEC int opal_mem_hooks_finalize(void);
-
-
 /**
  * Query level of support provided by memory hooks
  *
@@ -86,15 +69,15 @@ OPAL_DECLSPEC int opal_mem_hooks_finalize(void);
  * support is provided or a bit-wise OR of the available return values
  * if support is provided.
  *
- * @retval OPAL_MEMORY_FREE_SUPPORT   Memory hooks subsytem can trigger
- *                                    callback events when memory is going 
+ * @retval OPAL_MEMORY_FREE_SUPPORT   Memory hooks subsystem can trigger
+ *                                    callback events when memory is going
  *                                    to be released by the process, either
  *                                    by the user calling an allocator
  *                                    function or munmap.  Implies
  *                                    OPAL_MEMORY_MUNMAP_SUPPORT.
  * @retval OPAL_MEMORY_MUNMAP_SUPPORT Subsystem can trigger callback events
  *                                    by the user calling munmap directly.
- * @retval OPAL_MEMORY_CHUNK_SUPPORT  Memory hooks subsystem will only 
+ * @retval OPAL_MEMORY_CHUNK_SUPPORT  Memory hooks subsystem will only
  *                                    trigger callback events when the
  *                                    process is giving memory back to the
  *                                    operating system, not at ever call
@@ -104,7 +87,6 @@ OPAL_DECLSPEC int opal_mem_hooks_finalize(void);
  */
 OPAL_DECLSPEC int opal_mem_hooks_support_level(void);
 
-
 /**
  * Memory status change callback
  *
@@ -113,17 +95,15 @@ OPAL_DECLSPEC int opal_mem_hooks_support_level(void);
  * the note in opal_mem_hooks_register_alloc() or
  * opal_mem_hooks_register_release().
  *
- * @param buf     Pointer to the start of the allocation 
- * @param lentgh  Length of the allocation
+ * @param buf     Pointer to the start of the allocation
+ * @param length  Length of the allocation
  * @param cbdata  Data passed to memory hooks when callback
  *                was registered
  * @param from_alloc True if the callback is caused by a call to the
  *                general allocation routines (malloc, calloc, free,
  *                etc.) or directly from the user (mmap, munmap, etc.)
  */
-typedef void (opal_mem_hooks_callback_fn_t)(void *buf, size_t length, 
-                                            void *cbdata, bool from_alloc);
-
+typedef void(opal_mem_hooks_callback_fn_t)(void *buf, size_t length, void *cbdata, bool from_alloc);
 
 /**
  * Register callback for when memory is to be released
@@ -138,11 +118,10 @@ typedef void (opal_mem_hooks_callback_fn_t)(void *buf, size_t length,
  * @retval OPAL_SUCCESS The registration completed successfully.
  * @retval OPAL_EXISTS  The function is already registered and will not
  *                      be registered again.
- * @retval OPAL_ERR_NOT_SUPPORTED There are no hooks available for 
+ * @retval OPAL_ERR_NOT_SUPPORTED There are no hooks available for
  *                      receiving callbacks when memory is to be released
  */
-OPAL_DECLSPEC int opal_mem_hooks_register_release(opal_mem_hooks_callback_fn_t *func, 
-                                                  void *cbdata);
+OPAL_DECLSPEC int opal_mem_hooks_register_release(opal_mem_hooks_callback_fn_t *func, void *cbdata);
 
 /**
  * Unregister previously registered release callback
@@ -156,7 +135,6 @@ OPAL_DECLSPEC int opal_mem_hooks_register_release(opal_mem_hooks_callback_fn_t *
  */
 OPAL_DECLSPEC int opal_mem_hooks_unregister_release(opal_mem_hooks_callback_fn_t *func);
 
-
 END_C_DECLS
 
 #endif /* OPAL_MEMORY_MEMORY_H */
